/**
 * @fileoverview
 *
 * ActivityStreamFetcher is a class for periodically downloading activity
 * streams from an arbitrary number or URLs.
 *
 * An example:
 *
 * var asf = new ActivityStreamFetcher();
 * asf.addCallback(onStreamUpdated);
 * asf.addErrback(onStreamError);
 * asd.addStream(streamUrl1)
 * asd.addStream(streamUrl2)
 * asd.addStream(streamUrl3)
 * 
 * When any of the streams are downloaded the onStreamUpdated function will
 * be called with two arguments: the url and the activity stream data
 *
 * It's possible to add more than one success/error callback. All will be
 * called, but there are no guarantees they will be called in the same order
 * they were added!
 *
 * It's also possible to add success/error callbacks for each url. This can
 * be done either in the call to addStream:
 *
 * asf.addStream(streanUrl4, onUrl4Success, onUrl4Fail);
 *
 * or by providing a url argument to the addCallback/addErrback call:
 *
 * asf.addCallback(streamUrl3)
 *
 */
function ActivityStreamFetcher()
{

    // Make sure our asparser dependency is met
    try {
        new ActivityStreamParser();
    }
    catch (e)
    {
        throw "ActivityStreamParser class could not be found!\n" +
              "remember to add a script tag for activitystreamfetcher.js " +
              "to the html document";
    }
    
    /**
     * Streams are objects with the shape
     *  {
     *    url: <url>,
     *    updateTime: <Date object for most recent update>
     *  }
     *  
    */
    this._streams = [];

    /**
     * Callbacks: [{url, callback}..]
     * If url is None it's for all urls.
     */
    this._callbacks = [];
    this._errbacks = [];
    this._timerId = null;
    this.CRON_INTERVAL = 2500; // in ms. How often cron fun will run to check if there is work to do.
    this.REQUEST_INTERVAL = 1000 * 60 * 5; // in ms. Time between refresh
    this.MAX_REQUST_TIME = 1000 * 30; // in ms. How long to wait before a request is assumed to have died.
    this.MAX_RETRIES = 5;

    /**
     * @private
     * @constructor
     */
    this._init = function()
    {
        var self = this;
        var fun = function() {
            self._cron();
        }
        window.setInterval(fun, this.CRON_INTERVAL);
    }

    /**
     * Add a stream that will be polled periodically
     * @argument {string} url url for the stream
     * @argument {function} callback function called when stream has been downloaded. OPTIONAL
     * @argument {function} errback function called if the stream could not be downloaded. OPTIONAL
     */
    this.addStream = function(url, callback, errback)
    {
        this._streams.push({url: url, updated: null, retries: 0});
        if (callback) { this.addCallback(callback, url); }
        if (errback) { this.addErrback(errback, url); }
    }
    
    this.removeStream = function(url)
    {
        var ffun = function(e) { return e.url != url };
        this._streams = this._streams.filter(ffun);
        this._callbacks = this._callbacks.filter(ffun);
        this._errbacks = this._errbacks.filter(ffun);
    }
    
    /**
     * Add a callback for one or every stream url
     * @argument {function} callback the callback called when stream has been downloaded
     * @argument {string} url the url for which the callback is valid. If not given, callback applies to _all_ streams
     */
    this.addCallback = function(callback, url)
    {
        this._callbacks.push({callback: callback, url: url || null});
    }

    this.addErrback = function(errback, url)
    {
        this._errbacks.push({callback: errback, url: url || null})
    }
    
    /**
     * Force a refresh of all streams
     */
    this.refreshStreams = function()
    {
        // just clear updated flags. The next cron will do the rest.
        this._streams.forEach(function(e) {
            e.updated = null;
        });
    }

    /**
     * Periodic task that makes stuff happen..
     *  - checks if there are stale requests
     *  - starts next pending request.
     * @private
     */
    this._cron = function()
    {
        if (this._activeConn)
        {
            var dur = new Date().getTime() - this._activeConn.startTime;
            if (dur < this.MAX_REQUST_TIME)
            {
                // Something is in progress and not done yet. Nothing
                // for us to do them
                return;
            }
            else
            {
                // A connection seems to have hung
                this._onFeedFailed();
            }
        }

        var self = this;
        for (var n=0, stream; stream = this._streams[n]; n++)
        {
           if (stream.updated==null ||
                (new Date().getTime() - stream.updated) > self.REQUEST_INTERVAL)
            {
                self._startRequest(stream);
                return;
            }
        }
    }
    
    /**
     * @private
     */
    this._onFeedDownloaded = function(conn)
    {
        var parser = new ActivityStreamParser();
        var data = parser.parseDocument(conn.responseXML);
    
        // filter function for grabbing all callbacks that apply.
        var ffun = function(e) { return e.url==null || e.url==conn.stream.url; }

        conn.stream.updated = new Date().getTime();

        this._activeConn = null;
        this._callbacks.filter(ffun).forEach(function(cb) {
            cb.callback(conn.stream.url, data);
        });
    }
    
    this._onFeedFailed = function()
    {
        var conn = this._activeConn;
        this._activeConn = null;
        if (! conn) // shouldn't happen, but may if there was an exception
        {
            opera.postError("An unknown error occured");
            return;
        }
        var stream = conn.stream;
        stream.updated = new Date().getTime();
        stream.retries++;
        
        if (stream.retries>=this.MAX_RETRIES)
        {
            var ffun = function(e) { return e.url==null || e.url==conn.stream.url; }
            this._errbacks.filter(ffun).forEach(function(eb) {
                eb.callback(stream.url);
            });
            this.removeStream(stream.url);
        }
    }
    
    /**
     * @private
     */
    this._startRequest = function(stream)
    {
        var conn = new XMLHttpRequest();
        conn.stream = stream;
        conn.startTime = new Date().getTime();
        conn.open("GET", stream.url);

        var self = this;
        var rsfun = function()
        {
            if (conn.readyState == 4)
            {
                if (conn.status == 200)
                {
                    self._onFeedDownloaded(conn);
                }
                else
                {
                    self._onFeedFailed(conn);
                }
            }
        }
        conn.onreadystatechange = rsfun;
        this._activeConn = conn;

        // This is a workaround for some implementations raising security
        // violation instead of following the spec wrgt requests for hosts
        // that can't be found.
        try
        {
            conn.send();
        }
        catch (e)
        {
            this._onFeedFailed();
        }

    }
    
    this.toString = function()
    {
        return "[ActivityStreamFetcher instance]";
    }
    
    this._init(); // invoke constructor!
}